/* Copyright (c) 1999 TTK Ciar / Bill Moyer
 * All rights reserved.
 *
 * This code is derived from code developed by TTK Ciar and Bill Moyer.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 *
 *  1. Redistributions of source code must retain the above copyright 
 *     notice, this list of conditions and the following disclaimer.
 *  2. The name of the author may not be used to endorse or promote 
 *     products derived from this software without specific prior 
 *     written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS 
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY 
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
 * IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR 
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 * IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <time.h>
#include <sys/utsname.h>

#define STDTYPES_VER "100"

int len_char;
int len_short;
int len_long;
int len_int;
int len_long_long;
int len_addr;
int len_fptr;
struct utsname stuff;
int eflavor;

int main() {
  int   i;
  char  cptr[9];
  char  achar;
  short ashort;
  long  along;
  int   aint;
  long long along_long;
  char *aaddr;
  FILE *filvar;
  int   l8;
  int   l16;
  int   l32;
  int   l64;
  int   l128;
  int   nkludge = 0;
  int (*fptr)(int);

  uname ( &stuff );

  l8 = 0;  l16 = 0;  l32 = 0;  l64 = 0;  l128 = 0;
  achar = 1; 
  for ( len_char  = 0; achar;  len_char++  ) achar  <<= 1;
  len_short = sizeof(ashort) * len_char;
  len_long  = sizeof(along)  * len_char;
  len_int   = sizeof(aint)   * len_char;
  len_addr  = sizeof(aaddr)  * len_char;
  len_fptr  = sizeof(fptr)   * len_char;
  len_long_long = sizeof(along_long) * len_char;

  strcpy ( cptr, "abcdefghijklmnop" );  /* in case sizeof(int) == 16 */
  eflavor = ( (((int*)cptr)[0] & 255 ) != 'a' ); /* 1 iff big-endian */

  switch ( len_char ) { 
    case   8 : if ( !l8 )   l8   = 1; break;
    case  16 : if ( !l16 )  l16  = 1; break;
    case  32 : if ( !l32 )  l32  = 1; break;
    case  64 : if ( !l64 )  l64  = 1; break;
    case 128 : if ( !l128 ) l128 = 1; break;
    }

  switch ( len_short ) {
    case   8 : if ( !l8 )   l8   = 2; break;
    case  16 : if ( !l16 )  l16  = 2; break;
    case  32 : if ( !l32 )  l32  = 2; break;
    case  64 : if ( !l64 )  l64  = 2; break;
    case 128 : if ( !l128 ) l128 = 2; break;
    }

  switch ( len_int ) {
    case   8 : if ( !l8 )   l8   = 4; break;
    case  16 : if ( !l16 )  l16  = 4; break;
    case  32 : if ( !l32 )  l32  = 4; break;
    case  64 : if ( !l64 )  l64  = 4; break;
    case 128 : if ( !l128 ) l128 = 4; break;
    }

  switch ( len_long ) {
    case   8 : if ( !l8 )   l8   = 3; break;
    case  16 : if ( !l16 )  l16  = 3; break;
    case  32 : if ( !l32 )  l32  = 3; break;
    case  64 : if ( !l64 )  l64  = 3; break;
    case 128 : if ( !l128 ) l128 = 3; break;
    }

  switch ( len_long_long ) {
    case   8 : if ( !l8 )   l8   = 5; break;
    case  16 : if ( !l16 )  l16  = 5; break;
    case  32 : if ( !l32 )  l32  = 5; break;
    case  64 : if ( !l64 )  l64  = 5; break;
    case 128 : if ( !l128 ) l128 = 5; break;
    }

  printf ( "%s %s %s %s %s \n", stuff.sysname, stuff.nodename, 
           stuff.release, stuff.version, stuff.machine );
  printf ( "char  is %3d bits\n", len_char  ); 
  printf ( "short is %3d bits\n", len_short );
  printf ( "long  is %3d bits\n", len_long  );
  printf ( "int   is %3d bits\n", len_int   );
  printf ( "ptr   is %3d bits\n", len_addr  );
  printf ( "fptr  is %3d bits\n", len_fptr  );
  printf ( "long long is %3d bits\n", len_long_long );

  filvar = fopen ( "stdtypes.h", "w" );
  if ( filvar == NULL ) {
    printf ("stdtypes: cannot open file stdtypes.h, aborting\n");
    exit(1);
    }

  fprintf ( filvar, "/*  stdtypes.h   */\n" ); 
  fprintf ( filvar, "/* Copyright (c) 1999 TTK Ciar / Bill Moyer\n" );
  fprintf ( filvar, " * All rights reserved.\n" );
  fprintf ( filvar, " *\n" );
  fprintf ( filvar, " * This code is derived from code developed by TTK Ciar and Bill Moyer.\n" );
  fprintf ( filvar, " *\n" );
  fprintf ( filvar, " * Redistribution and use in source and binary forms, with or without \n" );
  fprintf ( filvar, " * modification, are permitted provided that the following conditions \n" );
  fprintf ( filvar, " * are met:\n" );
  fprintf ( filvar, " *\n" );
  fprintf ( filvar, " *  1. Redistributions of source code must retain the above copyright \n" );
  fprintf ( filvar, " *     notice, this list of conditions and the following disclaimer.\n" );
  fprintf ( filvar, " *  2. The name of the author may not be used to endorse or promote \n" );
  fprintf ( filvar, " *     products derived from this software without specific prior \n" );
  fprintf ( filvar, " *     written permission.\n" );
  fprintf ( filvar, " *\n" );
  fprintf ( filvar, " * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS \n" );
  fprintf ( filvar, " * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED \n" );
  fprintf ( filvar, " * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE \n" );
  fprintf ( filvar, " * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY \n" );
  fprintf ( filvar, " * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL \n" );
  fprintf ( filvar, " * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE\n" );
  fprintf ( filvar, " * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS \n" );
  fprintf ( filvar, " * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER\n" );
  fprintf ( filvar, " * IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR \n" );
  fprintf ( filvar, " * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN\n" );
  fprintf ( filvar, " * IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.\n" );
  fprintf ( filvar, " */\n\n" );

  fprintf ( filvar, "/*  This file automatically generated by stdtypes.c  */\n" );
  fprintf ( filvar, "/*  defines standard types of known lengths          */\n" );
  fprintf ( filvar, "/*  This file was generated on:                     \n" );
  fprintf ( filvar, "\t%s\n\t%s\n\t%s\n\t%s\n\t%s\n*/\n", 
     stuff.sysname, stuff.nodename, stuff.release, stuff.version, stuff.machine );
  fprintf ( filvar, "\n" );
  fprintf ( filvar, "#ifndef STDTYPES_VER\n");
  fprintf ( filvar, "#define STDTYPES_VER   %s\n", STDTYPES_VER );
  fprintf ( filvar, "#define LEN_CHAR       %d\n", len_char  );
  fprintf ( filvar, "#define LEN_SHORT      %d\n", len_short );
  fprintf ( filvar, "#define LEN_LONG       %d\n", len_long  );
  fprintf ( filvar, "#define LEN_INT        %d\n", len_int   );
  fprintf ( filvar, "#define LEN_PTR        %d\n", len_addr  );
  fprintf ( filvar, "#define LEN_FPTR       %d\n", len_fptr  );
  fprintf ( filvar, "#define LEN_LONG_LONG  %d\n", len_long_long );
  fprintf ( filvar, "\n");
  if ( eflavor ) 
       fprintf ( filvar, "#define BIGENDIAN\n");
  else fprintf ( filvar, "#define LITENDIAN\n");
  fprintf ( filvar, "\n");
  fprintf ( filvar, "typedef unsigned char      bool;\n");
  switch ( l8 ) {
    case 0 : fprintf ( filvar, "#define NO_8_BIT\n"); 
             fprintf ( filvar, "typedef unsigned char      byte;\n" ); break;
    case 1 : fprintf ( filvar, "typedef unsigned char      byte;\n" ); break;
    case 2 : fprintf ( filvar, "typedef unsigned short     byte;\n" ); break;
    case 3 : fprintf ( filvar, "typedef unsigned long      byte;\n" ); break;
    case 4 : fprintf ( filvar, "typedef unsigned int       byte;\n" ); break;
    case 5 : fprintf ( filvar, "typedef unsigned long long byte;\n" ); break;
    }
  switch ( l8 ) {
    case 0 : break;
    case 1 : fprintf ( filvar, "typedef unsigned char      word8;\n" ); break;
    case 2 : fprintf ( filvar, "typedef unsigned short     word8;\n" ); break;
    case 3 : fprintf ( filvar, "typedef unsigned long      word8;\n" ); break;
    case 4 : fprintf ( filvar, "typedef unsigned int       word8;\n" ); break;
    case 5 : fprintf ( filvar, "typedef unsigned long long word8;\n" ); break;
    }
  switch ( l16 ) {
    case 0 : fprintf ( filvar, "#define NO_16_BIT\n"); break;
    case 1 : fprintf ( filvar, "typedef unsigned char      word16;\n" ); break;
    case 2 : fprintf ( filvar, "typedef unsigned short     word16;\n" ); break;
    case 3 : fprintf ( filvar, "typedef unsigned long      word16;\n" ); break;
    case 4 : fprintf ( filvar, "typedef unsigned int       word16;\n" ); break;
    case 5 : fprintf ( filvar, "typedef unsigned long long word16;\n" ); break;
    }
  switch ( l32 ) {
    case 0 : fprintf ( filvar, "#define NO_32_BIT\n"); break;
    case 1 : fprintf ( filvar, "typedef unsigned char      word32;\n" ); break;
    case 2 : fprintf ( filvar, "typedef unsigned short     word32;\n" ); break;
    case 3 : fprintf ( filvar, "typedef unsigned long      word32;\n" ); break;
    case 4 : fprintf ( filvar, "typedef unsigned int       word32;\n" ); break;
    case 5 : fprintf ( filvar, "typedef unsigned long long word32;\n" ); break;
    }
  switch ( l64 ) {
    case 0 : fprintf ( filvar, "#define NO_64_BIT\n"); break;
    case 1 : fprintf ( filvar, "typedef unsigned char      word64;\n" ); break;
    case 2 : fprintf ( filvar, "typedef unsigned short     word64;\n" ); break;
    case 3 : fprintf ( filvar, "typedef unsigned long      word64;\n" ); break;
    case 4 : fprintf ( filvar, "typedef unsigned int       word64;\n" ); break;
    case 5 : fprintf ( filvar, "typedef unsigned long long word64;\n" ); break;
    }
  switch ( l128 ) {
    case 0 : fprintf ( filvar, "#define NO_128_BIT\n"); break;
    case 1 : fprintf ( filvar, "typedef unsigned char      word128;\n" ); break;
    case 2 : fprintf ( filvar, "typedef unsigned short     word128;\n" ); break;
    case 3 : fprintf ( filvar, "typedef unsigned long      word128;\n" ); break;
    case 4 : fprintf ( filvar, "typedef unsigned int       word128;\n" ); break;
    case 5 : fprintf ( filvar, "typedef unsigned long long word128;\n" ); break;
    }
  fprintf( filvar, "\n" );
  switch ( l8 ) {
    case 0 : break;
    case 1 : fprintf ( filvar, "typedef char      int8;\n" ); break;
    case 2 : fprintf ( filvar, "typedef short     int8;\n" ); break;
    case 3 : fprintf ( filvar, "typedef long      int8;\n" ); break;
    case 4 : fprintf ( filvar, "typedef int       int8;\n" ); break;
    case 5 : fprintf ( filvar, "typedef long long int8;\n" ); break;
    }
  switch ( l16 ) {
    case 0 : break;
    case 1 : fprintf ( filvar, "typedef char      int16;\n" ); break;
    case 2 : fprintf ( filvar, "typedef short     int16;\n" ); break;
    case 3 : fprintf ( filvar, "typedef long      int16;\n" ); break;
    case 4 : fprintf ( filvar, "typedef int       int16;\n" ); break;
    case 5 : fprintf ( filvar, "typedef long long int16;\n" ); break;
    }
  switch ( l32 ) {
    case 0 : break;
    case 1 : fprintf ( filvar, "typedef char      int32;\n" ); break;
    case 2 : fprintf ( filvar, "typedef short     int32;\n" ); break;
    case 3 : fprintf ( filvar, "typedef long      int32;\n" ); break;
    case 4 : fprintf ( filvar, "typedef int       int32;\n" ); break;
    case 5 : fprintf ( filvar, "typedef long long int32;\n" ); break;
    }
  switch ( l64 ) {
    case 0 : break;
    case 1 : fprintf ( filvar, "typedef char      int64;\n" ); break;
    case 2 : fprintf ( filvar, "typedef short     int64;\n" ); break;
    case 3 : fprintf ( filvar, "typedef long      int64;\n" ); break;
    case 4 : fprintf ( filvar, "typedef int       int64;\n" ); break;
    case 5 : fprintf ( filvar, "typedef long long int64;\n" ); break;
    }
  switch ( l128 ) {
    case 0 : break;
    case 1 : fprintf ( filvar, "typedef char      int128;\n" ); break;
    case 2 : fprintf ( filvar, "typedef short     int128;\n" ); break;
    case 3 : fprintf ( filvar, "typedef long      int128;\n" ); break;
    case 4 : fprintf ( filvar, "typedef int       int128;\n" ); break;
    case 5 : fprintf ( filvar, "typedef long long int128;\n" ); break;
    }

  fprintf ( filvar, "\n" );

  switch ( len_addr ) {
    case   8 : fprintf ( filvar, "typedef int8      ptrint;\n" );
               fprintf ( filvar, "typedef word8     ptrword;\n" ); break;
    case  16 : fprintf ( filvar, "typedef int16     ptrint;\n" ); 
               fprintf ( filvar, "typedef word16    ptrword;\n" ); break;
    case  32 : fprintf ( filvar, "typedef int32     ptrint;\n" ); 
               fprintf ( filvar, "typedef word32    ptrword;\n" ); break;
    case  64 : fprintf ( filvar, "typedef int64     ptrint;\n" ); 
               fprintf ( filvar, "typedef word64    ptrword;\n" ); break;
    case 128 : fprintf ( filvar, "typedef int128    ptrint;\n" ); 
               fprintf ( filvar, "typedef word128   ptrword;\n" ); break;
    default:
         nkludge = len_addr / len_char;
         fprintf (filvar,"#define KLUDGE_ADDR\n");
         fprintf (filvar,"typedef struct intp_kludgetag {\n");
         for ( i = 0; i < nkludge; i++ ) 
           fprintf (filvar,"\tint8 byte%d;",i);
         fprintf (filvar,"\t} ptrint;");
         fprintf (filvar,"typedef struct wordp_kludgetag {\n");
         for ( i = 0; i < nkludge; i++ ) 
           fprintf (filvar,"\tword8 byte%d;",i);
         fprintf (filvar,"\t} ptrword;");
         break;
    } /* end of switch */

  switch ( len_fptr ) {
    case   8 : fprintf ( filvar, "typedef int8      fptrint;\n" ); 
               fprintf ( filvar, "typedef word8     fptrword;\n" ); break;
    case  16 : fprintf ( filvar, "typedef int16     fptrint;\n" ); 
               fprintf ( filvar, "typedef word16    fptrword;\n" ); break;
    case  32 : fprintf ( filvar, "typedef int32     fptrint;\n" ); 
               fprintf ( filvar, "typedef word32    fptrword;\n" ); break;
    case  64 : fprintf ( filvar, "typedef int64     fptrint;\n" ); 
               fprintf ( filvar, "typedef word64    fptrword;\n" ); break;
    case 128 : fprintf ( filvar, "typedef int128    fptrint;\n" ); 
               fprintf ( filvar, "typedef word128   fptrword;\n" ); break;
    default:
         nkludge = len_addr / len_char;
         fprintf (filvar,"#define KLUDGE_FADDR\n");
         fprintf (filvar,"typedef struct intfp_kludgetag {\n");
         for ( i = 0; i < nkludge; i++ ) 
           fprintf (filvar,"\tint8 byte%d;",i);
         fprintf (filvar,"\t} fptrint;");
         fprintf (filvar,"typedef struct wordfp_kludgetag {\n");
         for ( i = 0; i < nkludge; i++ ) 
           fprintf (filvar,"\tword8 byte%d;",i);
         fprintf (filvar,"\t} fptrword;");
         break;
    } /* end of switch */

  fprintf ( filvar, "\n#endif\n");
  fclose  ( filvar );
  exit(0);
  }
